from collections import Iterable

from pyschemes import Scheme, validators

from cg.constants import (
    CONTAINER_OPTIONS,
    PRIORITY_OPTIONS,
    SEX_OPTIONS,
    STATUS_OPTIONS,
    Pipeline,
)
from cg.utils.StrEnum import StrEnum


class OrderType(StrEnum):
    BALSAMIC: str = str(Pipeline.BALSAMIC)
    FASTQ: str = str(Pipeline.FASTQ)
    FLUFFY: str = str(Pipeline.FLUFFY)
    METAGENOME: str = "metagenome"
    MICROSALT: str = str(Pipeline.MICROSALT)
    MIP_DNA: str = str(Pipeline.MIP_DNA)
    MIP_RNA: str = str(Pipeline.MIP_RNA)
    RML: str = "rml"
    SARS_COV_2: str = str(Pipeline.SARS_COV_2)


class ListValidator(validators.Validator):

    """Validate a list of items against a schema."""

    def __init__(self, scheme: validators.Validator, min_items: int = 0):
        super().__init__(scheme)
        self.__scheme = validators.Validator.create_validator(scheme)
        self.min_items = min_items

    def validate(self, value: Iterable):
        """Validate the list of values."""
        values = value
        validators.TypeValidator(Iterable).validate(values)
        values_copy = []
        for one_value in values:
            self.__scheme.validate(one_value)
            values_copy.append(one_value)
        if len(values_copy) < self.min_items:
            raise ValueError(f"value did not contain at least {self.min_items} items")
        return values_copy


class TypeValidatorNone(validators.TypeValidator):
    """Type Scheme that accepts None."""

    def validate(self, value):
        """Validate the type of the value, accept None"""
        if value in (None, ""):
            return value
        return super().validate(value)


class RegexValidatorNone(validators.RegexValidator):
    """Validator for regex patterns that accepts None."""

    def validate(self, value):
        """Validate the the value against an regular expression, accept None"""
        if value in (None, ""):
            return value
        return super().validate(value)


class OptionalNone(validators.Optional):
    """Optional Scheme that accepts None."""

    def __init__(self, scheme, default=None):
        super().__init__(scheme, default)

    def validate(self, *value):
        """Validate the value with the given validator, accept missing values, accept None"""
        if value:
            val = value[0]

            if val in (None, ""):
                return val

            return super().validate(val)
        return self.default


NAME_PATTERN = r"^[A-Za-z0-9-]*$"

BASE_PROJECT = {"name": str, "customer": str, "comment": OptionalNone(TypeValidatorNone(str))}

MIP_SAMPLE = {
    # Orderform 1508
    # Order portal specific
    "internal_id": OptionalNone(TypeValidatorNone(str)),
    # "required for new samples"
    "name": validators.RegexValidator(NAME_PATTERN),
    # customer
    "data_analysis": str,
    "data_delivery": OptionalNone(TypeValidatorNone(str)),
    "age_at_sampling": OptionalNone(TypeValidatorNone(str)),
    "application": str,
    "family_name": validators.RegexValidator(NAME_PATTERN),
    "case_internal_id": OptionalNone(TypeValidatorNone(str)),
    "sex": OptionalNone(validators.Any(SEX_OPTIONS)),
    "tumour": bool,
    "source": OptionalNone(TypeValidatorNone(str)),
    "priority": OptionalNone(validators.Any(PRIORITY_OPTIONS)),
    "volume": OptionalNone(TypeValidatorNone(str)),
    "container": OptionalNone(validators.Any(CONTAINER_OPTIONS)),
    # "required if plate for new samples"
    "container_name": OptionalNone(TypeValidatorNone(str)),
    "well_position": OptionalNone(TypeValidatorNone(str)),
    # "Required if data analysis in Scout or vcf delivery"
    "panels": ListValidator(str, min_items=1),
    "status": OptionalNone(validators.Any(STATUS_OPTIONS)),
    # "Required if samples are part of trio/family"
    "mother": OptionalNone(RegexValidatorNone(NAME_PATTERN)),
    "father": OptionalNone(RegexValidatorNone(NAME_PATTERN)),
    # This information is required for panel analysis
    "capture_kit": OptionalNone(TypeValidatorNone(str)),
    # This information is required for panel- or exome analysis
    "elution_buffer": OptionalNone(TypeValidatorNone(str)),
    "tumour_purity": OptionalNone(TypeValidatorNone(str)),
    # "This information is optional for FFPE-samples for new samples"
    "formalin_fixation_time": OptionalNone(TypeValidatorNone(str)),
    "post_formalin_fixation_time": OptionalNone(TypeValidatorNone(str)),
    "tissue_block_size": OptionalNone(TypeValidatorNone(str)),
    # "Not Required"
    "quantity": OptionalNone(TypeValidatorNone(str)),
    "comment": OptionalNone(TypeValidatorNone(str)),
    "cohorts": OptionalNone(ListValidator(str, min_items=0)),
    "synopsis": OptionalNone(str),
    "subject_id": OptionalNone(validators.RegexValidator(NAME_PATTERN)),
    "phenotype_terms": OptionalNone(ListValidator(str, min_items=0)),
    "phenotype_groups": OptionalNone(ListValidator(str, min_items=0)),
}

BALSAMIC_SAMPLE = {
    # 1508 Orderform
    # Order portal specific
    "internal_id": OptionalNone(TypeValidatorNone(str)),
    # "This information is required for new samples"
    "name": validators.RegexValidator(NAME_PATTERN),
    "container": OptionalNone(validators.Any(CONTAINER_OPTIONS)),
    "data_analysis": str,
    "data_delivery": OptionalNone(TypeValidatorNone(str)),
    "application": str,
    "sex": OptionalNone(validators.Any(SEX_OPTIONS)),
    "family_name": validators.RegexValidator(NAME_PATTERN),
    "case_internal_id": OptionalNone(TypeValidatorNone(str)),
    "volume": OptionalNone(TypeValidatorNone(str)),
    "tumour": bool,
    "source": OptionalNone(TypeValidatorNone(str)),
    "priority": OptionalNone(validators.Any(PRIORITY_OPTIONS)),
    # Required if Plate for new samples
    "container_name": OptionalNone(TypeValidatorNone(str)),
    "well_position": OptionalNone(TypeValidatorNone(str)),
    # This information is required for panel analysis
    "capture_kit": OptionalNone(TypeValidatorNone(str)),
    # This information is required for panel- or exome analysis
    "elution_buffer": OptionalNone(TypeValidatorNone(str)),
    "tumour_purity": OptionalNone(TypeValidatorNone(str)),
    # This information is optional for FFPE-samples for new samples
    "formalin_fixation_time": OptionalNone(TypeValidatorNone(str)),
    "post_formalin_fixation_time": OptionalNone(TypeValidatorNone(str)),
    "tissue_block_size": OptionalNone(TypeValidatorNone(str)),
    # This information is optional
    "quantity": OptionalNone(TypeValidatorNone(str)),
    "comment": OptionalNone(TypeValidatorNone(str)),
    "age_at_sampling": OptionalNone(TypeValidatorNone(str)),
    "cohorts": OptionalNone(ListValidator(str, min_items=0)),
    "subject_id": OptionalNone(str),
    "synopsis": OptionalNone(validators.RegexValidator(NAME_PATTERN)),
    "phenotype_terms": OptionalNone(ListValidator(str, min_items=0)),
    "phenotype_groups": OptionalNone(ListValidator(str, min_items=0)),
}

MIP_RNA_SAMPLE = {
    "internal_id": OptionalNone(TypeValidatorNone(str)),
    # "required for new samples"
    "name": validators.RegexValidator(NAME_PATTERN),
    # customer
    "data_analysis": str,
    "data_delivery": OptionalNone(TypeValidatorNone(str)),
    "application": str,
    "family_name": validators.RegexValidator(NAME_PATTERN),
    "case_internal_id": OptionalNone(TypeValidatorNone(str)),
    "sex": OptionalNone(validators.Any(SEX_OPTIONS)),
    "source": OptionalNone(TypeValidatorNone(str)),
    "priority": OptionalNone(validators.Any(PRIORITY_OPTIONS)),
    "volume": OptionalNone(TypeValidatorNone(str)),
    "container": OptionalNone(validators.Any(CONTAINER_OPTIONS)),
    # "required if plate for new samples"
    "container_name": OptionalNone(TypeValidatorNone(str)),
    "well_position": OptionalNone(TypeValidatorNone(str)),
    "formalin_fixation_time": OptionalNone(TypeValidatorNone(str)),
    "post_formalin_fixation_time": OptionalNone(TypeValidatorNone(str)),
    "tissue_block_size": OptionalNone(TypeValidatorNone(str)),
    # # "Not Required"
    "quantity": OptionalNone(TypeValidatorNone(str)),
    "comment": OptionalNone(TypeValidatorNone(str)),
    "time_point": OptionalNone(TypeValidatorNone(str)),
    "age_at_sampling": OptionalNone(TypeValidatorNone(str)),
    "cohorts": OptionalNone(ListValidator(str, min_items=0)),
    "subject_id": OptionalNone(validators.RegexValidator(NAME_PATTERN)),
    "synopsis": OptionalNone(str),
    "phenotype_terms": OptionalNone(ListValidator(str, min_items=0)),
    "phenotype_groups": OptionalNone(ListValidator(str, min_items=0)),
}

FASTQ_SAMPLE = {
    # Orderform 1508
    # "required"
    "name": validators.RegexValidator(NAME_PATTERN),
    "container": OptionalNone(validators.Any(CONTAINER_OPTIONS)),
    "data_analysis": str,
    "data_delivery": OptionalNone(TypeValidatorNone(str)),
    "application": str,
    "sex": OptionalNone(validators.Any(SEX_OPTIONS)),
    "volume": str,
    "source": str,
    "tumour": bool,
    "priority": OptionalNone(validators.Any(PRIORITY_OPTIONS)),
    # "required if plate"
    "container_name": OptionalNone(TypeValidatorNone(str)),
    "well_position": OptionalNone(TypeValidatorNone(str)),
    # "Required if data analysis in Scout or vcf delivery" => not valid for fastq
    # 'panels': ListValidator(str, min_items=1),
    # 'status': OptionalNone(validators.Any(STATUS_OPTIONS),
    "elution_buffer": str,
    # "Required if samples are part of trio/family"
    "mother": OptionalNone(RegexValidatorNone(NAME_PATTERN)),
    "father": OptionalNone(RegexValidatorNone(NAME_PATTERN)),
    # "Not Required"
    "quantity": OptionalNone(TypeValidatorNone(str)),
    "comment": OptionalNone(TypeValidatorNone(str)),
}

RML_SAMPLE = {
    # 1604 Orderform Ready made libraries (RML)
    # Order portal specific
    "priority": str,
    # "This information is required"
    "name": validators.RegexValidator(NAME_PATTERN),
    "pool": str,
    "application": str,
    "data_analysis": str,
    "data_delivery": OptionalNone(TypeValidatorNone(str)),
    "volume": str,
    "concentration": str,
    "concentration_sample": OptionalNone(TypeValidatorNone(str)),
    "index": str,
    "index_number": OptionalNone(TypeValidatorNone(str)),  # optional for NoIndex
    # "Required if Plate"
    "rml_plate_name": OptionalNone(TypeValidatorNone(str)),
    "well_position_rml": OptionalNone(TypeValidatorNone(str)),
    # "Automatically generated (if not custom) or custom"
    "index_sequence": OptionalNone(TypeValidatorNone(str)),
    # "Not required"
    "comment": OptionalNone(TypeValidatorNone(str)),
    "control": OptionalNone(TypeValidatorNone(str)),
}

MICROSALT_SAMPLE = {
    # 1603 Orderform Microbial WGS
    # "These fields are required"
    "name": validators.RegexValidator(NAME_PATTERN),
    "organism": str,
    "reference_genome": str,
    "data_analysis": str,
    "data_delivery": str,
    "application": str,
    "priority": validators.Any(PRIORITY_OPTIONS),
    "require_qcok": bool,
    "elution_buffer": str,
    "extraction_method": str,
    "volume": str,
    "container": validators.Any(CONTAINER_OPTIONS),
    # "Required if Plate"
    "container_name": OptionalNone(TypeValidatorNone(str)),
    "well_position": OptionalNone(TypeValidatorNone(str)),
    # "Required if "Other" is chosen in column "Species""
    "organism_other": OptionalNone(TypeValidatorNone(str)),
    # "These fields are not required"
    "concentration_sample": OptionalNone(TypeValidatorNone(str)),
    "quantity": OptionalNone(TypeValidatorNone(str)),
    "comment": OptionalNone(TypeValidatorNone(str)),
}

METAGENOME_SAMPLE = {
    # 1605 Orderform Microbial Metagenomes- 16S
    # "This information is required"
    "name": validators.RegexValidator(NAME_PATTERN),
    "container": OptionalNone(validators.Any(CONTAINER_OPTIONS)),
    "data_analysis": str,
    "data_delivery": OptionalNone(TypeValidatorNone(str)),
    "application": str,
    "require_qcok": bool,
    "elution_buffer": str,
    "source": str,
    "volume": str,
    "priority": OptionalNone(validators.Any(PRIORITY_OPTIONS)),
    # "Required if Plate"
    "container_name": OptionalNone(TypeValidatorNone(str)),
    "well_position": OptionalNone(TypeValidatorNone(str)),
    # "This information is not required"
    "concentration_sample": OptionalNone(TypeValidatorNone(str)),
    "quantity": OptionalNone(TypeValidatorNone(str)),
    "extraction_method": OptionalNone(TypeValidatorNone(str)),
    "comment": OptionalNone(TypeValidatorNone(str)),
}

SARSCOV2_SAMPLE = {
    # 2184 Orderform SARS-COV-2
    # "These fields are required"
    "application": str,
    "collection_date": str,
    "container": validators.Any(CONTAINER_OPTIONS),
    "data_analysis": str,
    "data_delivery": str,
    "elution_buffer": str,
    "extraction_method": str,
    "lab_code": str,
    "name": validators.RegexValidator(NAME_PATTERN),
    "organism": str,
    "original_lab": str,
    "original_lab_address": str,
    "pre_processing_method": str,
    "priority": validators.Any(PRIORITY_OPTIONS),
    "reference_genome": str,
    "region": str,
    "region_code": str,
    "require_qcok": bool,
    "selection_criteria": str,
    "volume": str,
    # "Required if Plate"
    "container_name": OptionalNone(TypeValidatorNone(str)),
    "well_position": OptionalNone(TypeValidatorNone(str)),
    # "Required if "Other" is chosen in column "Species""
    "organism_other": OptionalNone(TypeValidatorNone(str)),
    # "These fields are not required"
    "comment": OptionalNone(TypeValidatorNone(str)),
    "concentration_sample": OptionalNone(TypeValidatorNone(str)),
    "quantity": OptionalNone(TypeValidatorNone(str)),
}

ORDER_SCHEMES = {
    OrderType.MIP_DNA: Scheme({**BASE_PROJECT, "samples": ListValidator(MIP_SAMPLE, min_items=1)}),
    OrderType.BALSAMIC: Scheme(
        {**BASE_PROJECT, "samples": ListValidator(BALSAMIC_SAMPLE, min_items=1)}
    ),
    OrderType.MIP_RNA: Scheme(
        {**BASE_PROJECT, "samples": ListValidator(MIP_RNA_SAMPLE, min_items=1)}
    ),
    OrderType.FASTQ: Scheme({**BASE_PROJECT, "samples": ListValidator(FASTQ_SAMPLE, min_items=1)}),
    OrderType.RML: Scheme({**BASE_PROJECT, "samples": ListValidator(RML_SAMPLE, min_items=1)}),
    OrderType.FLUFFY: Scheme({**BASE_PROJECT, "samples": ListValidator(RML_SAMPLE, min_items=1)}),
    OrderType.MICROSALT: Scheme(
        {**BASE_PROJECT, "samples": ListValidator(MICROSALT_SAMPLE, min_items=1)}
    ),
    OrderType.METAGENOME: Scheme(
        {**BASE_PROJECT, "samples": ListValidator(METAGENOME_SAMPLE, min_items=1)}
    ),
    OrderType.SARS_COV_2: Scheme(
        {**BASE_PROJECT, "samples": ListValidator(SARSCOV2_SAMPLE, min_items=1)}
    ),
}



from typing import Any, Dict, List, Optional

from pydantic import validator

from cg.constants.gender import Gender
from cg.models.deliverables.metric_deliverables import (
    SampleMetric,
    MedianTargetCoverage,
    ParsedMetrics,
    MeanInsertSize,
    MetricsDeliverables,
    add_metric,
    add_sample_id_metrics,
)

SAMPLE_METRICS_TO_PARSE: List[str] = [
    "duplicate_reads",
    "mapped_reads",
    "mean_insert_size",
    "median_target_coverage",
    "predicted_sex",
]


class DuplicateReads(SampleMetric):
    """Definition of duplicate reads metric"""

    value: float

    @validator("value", always=True)
    def convert_duplicate_read(cls, value) -> float:
        """Convert raw value from fraction to percent"""
        return value * 100


class GenderCheck(SampleMetric):
    """Definition of gender check metric"""

    value: str


class MIPMappedReads(SampleMetric):
    """Definition of mapped reads metric"""

    value: float

    @validator("value", always=True)
    def convert_mapped_read(cls, value) -> float:
        """Convert raw value from fraction to percent"""
        return value * 100


class MIPParsedMetrics(ParsedMetrics):
    """Defines parsed metrics"""

    duplicate_reads: float
    duplicate_reads_step: str
    mapped_reads: float
    mapped_reads_step: str
    predicted_sex: str = Gender.UNKNOWN
    predicted_sex_step: str


class MIPMetricsDeliverables(MetricsDeliverables):
    """Specification for a metric MIP deliverables file"""

    metric_to_get_: Dict[str, Any] = {
        "fraction_duplicates": DuplicateReads,
        "MEAN_INSERT_SIZE": MeanInsertSize,
        "MEDIAN_TARGET_COVERAGE": MedianTargetCoverage,
        "gender": GenderCheck,
    }
    duplicate_reads: Optional[List[DuplicateReads]]
    mapped_reads: Optional[List[MIPMappedReads]]
    mean_insert_size: Optional[List[MeanInsertSize]]
    median_target_coverage: Optional[List[MedianTargetCoverage]]
    predicted_sex: Optional[List[GenderCheck]]
    sample_metric_to_parse: List[str] = SAMPLE_METRICS_TO_PARSE
    sample_id_metrics: Optional[List[MIPParsedMetrics]]

    @validator("duplicate_reads", always=True)
    def set_duplicate_reads(cls, _, values: dict) -> List[DuplicateReads]:
        """Set duplicate_reads"""
        return add_metric(name="fraction_duplicates", values=values)

    @validator("mapped_reads", always=True)
    def set_mapped_reads(cls, _, values: dict) -> List[MIPMappedReads]:
        """Set mapped reads"""
        sample_ids: set = values.get("sample_ids")
        mapped_reads: list = []
        total_sequences: dict = {}
        reads_mapped: dict = {}
        raw_metrics: list = values.get("metrics_")
        metric_step: str = ""
        for metric in raw_metrics:
            if metric.name == "raw_total_sequences":
                raw_total_sequences = total_sequences.get(metric.id, 0)
                total_sequences[metric.id] = int(metric.value) + raw_total_sequences
                metric_step: str = metric.step
            if metric.name == "reads_mapped":
                raw_reads_mapped = reads_mapped.get(metric.id, 0)
                reads_mapped[metric.id] = int(metric.value) + raw_reads_mapped
        for sample_id in sample_ids:
            fraction_mapped_read = reads_mapped[sample_id] / total_sequences[sample_id]
            mapped_reads.append(
                MIPMappedReads(sample_id=sample_id, step=metric_step, value=fraction_mapped_read)
            )
        return mapped_reads

    @validator("mean_insert_size", always=True)
    def set_mean_insert_size(cls, _, values: dict) -> List[MeanInsertSize]:
        """Set mean insert size"""
        return add_metric(name="MEAN_INSERT_SIZE", values=values)

    @validator("median_target_coverage", always=True)
    def set_median_target_coverage(cls, _, values: dict) -> List[MedianTargetCoverage]:
        """Set median target coverage"""
        return add_metric(name="MEDIAN_TARGET_COVERAGE", values=values)

    @validator("predicted_sex", always=True)
    def set_predicted_sex(cls, _, values: dict) -> List[GenderCheck]:
        """Set predicted sex"""
        return add_metric(name="gender", values=values)

    @validator("sample_id_metrics", always=True)
    def set_sample_id_metrics(cls, _, values: dict) -> List[MIPParsedMetrics]:
        """Set parsed sample_id metrics gathered from all metrics"""
        return add_sample_id_metrics(parsed_metric=MIPParsedMetrics, values=values)


def get_sample_id_metric(
    sample_id_metrics: List[MIPParsedMetrics], sample_id: str
) -> MIPParsedMetrics:
    """Get parsed metrics for an sample_id"""
    for sample_id_metric in sample_id_metrics:
        if sample_id == sample_id_metric.sample_id:
            return sample_id_metric



"""API for transfer a flowcell"""

import logging
from pathlib import Path
from typing import List

from cg.apps.cgstats.stats import StatsAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.models.cgstats.flowcell import StatsFlowcell
from cg.store import Store, models
from housekeeper.store import models as hk_models

LOG = logging.getLogger(__name__)


class TransferFlowcell:
    """Transfer flowcell API"""

    def __init__(self, db: Store, stats_api: StatsAPI, hk_api: HousekeeperAPI):
        self.db: Store = db
        self.stats: StatsAPI = stats_api
        self.hk: HousekeeperAPI = hk_api

    def transfer(self, flowcell_name: str, store: bool = True) -> models.Flowcell:
        """Populate the database with the information."""
        if store and self.hk.tag("fastq") is None:
            self.hk.add_commit(self.hk.new_tag("fastq"))
        if store and self.hk.tag("samplesheet") is None:
            self.hk.add_commit(self.hk.new_tag("samplesheet"))
        if store and self.hk.tag(flowcell_name) is None:
            self.hk.add_commit(self.hk.new_tag(flowcell_name))
        stats_data: StatsFlowcell = self.stats.flowcell(flowcell_name)
        flowcell_obj: models.Flowcell = self.db.flowcell(flowcell_name)

        if flowcell_obj is None:
            flowcell_obj: models.Flowcell = self.db.add_flowcell(
                name=flowcell_name,
                sequencer=stats_data.sequencer,
                sequencer_type=stats_data.sequencer_type,
                date=stats_data.date,
            )
        flowcell_obj.status = "ondisk"

        sample_sheet_path = self._sample_sheet_path(flowcell_name)
        if not Path(sample_sheet_path).exists():
            LOG.warning(f"unable to find samplesheet: {sample_sheet_path}")
        elif store:
            self.store_samplesheet(flowcell_name, sample_sheet_path)

        for sample_data in stats_data.samples:
            LOG.debug(f"adding reads/fastqs to sample: {sample_data.name}")
            sample_obj = self.db.sample(sample_data.name)
            if sample_obj is None:
                LOG.warning(f"unable to find sample: {sample_data.name}")
                continue

            if store:
                self.store_fastqs(
                    sample=sample_obj.internal_id,
                    flowcell=flowcell_name,
                    fastq_files=sample_data.fastqs,
                )

            sample_obj.reads = sample_data.reads
            enough_reads = (
                sample_obj.reads > sample_obj.application_version.application.expected_reads
            )
            newest_date = (sample_obj.sequenced_at is None) or (
                flowcell_obj.sequenced_at > sample_obj.sequenced_at
            )
            if newest_date:
                sample_obj.sequenced_at = flowcell_obj.sequenced_at

            if isinstance(sample_obj, models.Sample):
                flowcell_obj.samples.append(sample_obj)

            LOG.info(
                f"added reads to sample: {sample_data.name} - {sample_data.reads} "
                f"[{'DONE' if enough_reads else 'NOT DONE'}]"
            )

        return flowcell_obj

    def store_fastqs(self, sample: str, flowcell: str, fastq_files: List[str]):
        """Store FASTQ files for a sample in housekeeper."""
        hk_bundle = self.hk.bundle(sample)
        if hk_bundle is None:
            hk_bundle = self.hk.new_bundle(sample)
            self.hk.add_commit(hk_bundle)
            new_version = self.hk.new_version(created_at=hk_bundle.created_at)
            hk_bundle.versions.append(new_version)
            self.hk.commit()
            LOG.info(f"added new Housekeeper bundle: {hk_bundle.name}")

        with self.hk.session_no_autoflush():
            hk_version = hk_bundle.versions[0]
            for fastq_file in fastq_files:
                if self.hk.files(path=fastq_file).first() is None:
                    LOG.info(f"found FASTQ file: {fastq_file}")
                    tags = [self.hk.tag("fastq"), self.hk.tag(flowcell)]
                    new_file = self.hk.new_file(path=fastq_file, tags=tags)
                    hk_version.files.append(new_file)
            self.hk.commit()

    def store_samplesheet(self, flowcell: str, sample_sheet_path: str):
        """Store samplesheet for a run in Housekeeper"""
        hk_bundle = self.hk.bundle(flowcell)
        if hk_bundle is None:
            hk_bundle = self.hk.new_bundle(flowcell)
            self.hk.add_commit(hk_bundle)
            new_version = self.hk.new_version(created_at=hk_bundle.created_at)
            hk_bundle.versions.append(new_version)
            self.hk.commit()
            LOG.info(f"added new Housekeeper bundle: {hk_bundle.name}")

        with self.hk.session_no_autoflush():
            hk_version = hk_bundle.versions[0]
            if self.hk.files(path=sample_sheet_path).first() is None:
                LOG.info(f"Adding samplesheet: {sample_sheet_path}")
                tags = [self.hk.tag("samplesheet"), self.hk.tag(flowcell)]
                new_file = self.hk.new_file(path=sample_sheet_path, tags=tags)
                hk_version.files.append(new_file)
        self.hk.commit()

    def _sample_sheet_path(self, flowcell: str) -> str:
        """Construct the path to the samplesheet to be stored"""
        run_name: str = self.stats.run_name(flowcell)
        document_path: str = self.stats.document_path(flowcell)
        unaligned_dir: str = Path(document_path).name
        root_dir: Path = self.stats.root_dir
        return str(root_dir.joinpath(run_name, unaligned_dir, "SampleSheet.csv"))



"""
    conftest for genotype API
"""

import json

import pytest


@pytest.fixture(name="genotype_export_sample_output")
def fixture_genotype_export_sample_output(genotype_config: dict) -> str:
    """
    genotype API fixture
    """
    data = {
        "ACC6987A15": {
            "status": None,
            "sample_created_in_genotype_db": "2020-07-15",
            "sex": "female",
            "comment": None,
        },
        "ACC6987A16": {
            "status": None,
            "sample_created_in_genotype_db": "2020-07-15",
            "sex": "female",
            "comment": None,
        },
    }

    return json.dumps(data)


@pytest.fixture(name="genotype_export_sample_analysis_output")
def fixture_genotype_export_sample_analysis_output() -> str:
    """Return some output from a sample analysis export"""
    data = {
        "ACC6987A15": {
            "snps": {
                "sequence": {
                    "rs10144418": ["T", "T"],
                    "rs1037256": ["G", "A"],
                    "rs1044973": ["C", "T"],
                    "rs1065772": ["C", "T"],
                    "rs11010": ["T", "C"],
                    "rs11789987": ["C", "T"],
                    "rs11797": ["C", "C"],
                }
            }
        },
        "ACC6987A16": {
            "snps": {
                "sequence": {
                    "rs10144418": ["T", "C"],
                    "rs1037256": ["G", "A"],
                    "rs1044973": ["C", "T"],
                    "rs1065772": ["C", "C"],
                    "rs11010": ["T", "T"],
                    "rs11789987": ["C", "T"],
                    "rs11797": ["T", "T"],
                }
            }
        },
    }
    return json.dumps(data)



import logging
import shutil
from pathlib import Path

import pytest
from click.testing import CliRunner

from cg.apps.cgstats.stats import StatsAPI
from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.apps.lims import LimsAPI
from cg.models.cg_config import CGConfig, DemultiplexConfig
from cg.models.demultiplex.flowcell import Flowcell
from cg.utils import Process
from tests.apps.cgstats.conftest import (
    fixture_bcl2fastq_demux_results,
    fixture_demultiplexed_flowcell,
    fixture_demultiplexed_runs,
    fixture_populated_stats_api,
    fixture_stats_api,
)
from tests.apps.crunchy.conftest import fixture_sbatch_process
from tests.apps.demultiplex.conftest import (
    fixture_demultiplex_fixtures,
    fixture_demux_run_dir,
    fixture_demux_run_dir_bcl2fastq,
    fixture_demux_run_dir_dragen,
    fixture_flowcell_full_name,
    fixture_flowcell_name,
    fixture_lims_novaseq_bcl2fastq_samples,
    fixture_lims_novaseq_dragen_samples,
    fixture_lims_novaseq_samples,
    fixture_lims_novaseq_samples_file,
    fixture_novaseq_dir,
    fixture_novaseq_dir_bcl2fastq,
    fixture_novaseq_dir_dragen,
    fixture_novaseq_run_parameters,
    fixture_raw_samples_dir,
)

LOG = logging.getLogger(__name__)


@pytest.fixture(name="demux_results_finished_dir")
def fixture_demux_results_finished_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to a dir with demultiplexing results where files are renamed etc"""
    return demultiplex_fixtures / "demultiplexed-runs"


@pytest.fixture(name="demux_results_not_finished_dir")
def fixture_demux_results_not_finished_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to a dir with demultiplexing results where demux has been done but nothing is cleaned"""
    return demultiplex_fixtures / "demultiplexed-runs-unfinished"


@pytest.fixture(name="flowcell_object")
def fixture_flowcell_object(demux_run_dir: Path, flowcell_full_name: str) -> Flowcell:
    """Create a flowcell object with flowcell that is demultiplexed"""
    return Flowcell(flowcell_path=demux_run_dir / flowcell_full_name)


@pytest.fixture(name="novaseq_bcl2fastq_sample_sheet_path")
def fixture_novaseq_bcl2fastq_sample_sheet_path(demultiplex_fixtures: Path) -> Path:
    """Return the path to a novaseq bcl2fastq sample sheet"""
    return demultiplex_fixtures / "SampleSheetS2_Bcl2Fastq.csv"


@pytest.fixture(name="novaseq_dragen_sample_sheet_path")
def fixture_novaseq_dragen_sample_sheet_path(demultiplex_fixtures: Path) -> Path:
    """Return the path to a novaseq bcl2fastq sample sheet"""
    return demultiplex_fixtures / "SampleSheetS2_Dragen.csv"


@pytest.fixture(name="flowcell_runs_working_directory")
def fixture_flowcell_runs_working_directory(project_dir: Path):
    """Return the path to a working directory with flowcells ready for demux"""
    working_dir = project_dir / "flowcell-runs"
    working_dir.mkdir(parents=True)
    return working_dir


@pytest.fixture(name="flowcell_runs_working_directory_bcl2fastq")
def fixture_flowcell_runs_working_directory_bcl2fastq(flowcell_runs_working_directory: Path):
    """Return the path to a working directory with flowcells ready for demux"""
    working_dir = flowcell_runs_working_directory / "bcl2fastq"
    working_dir.mkdir(parents=True)
    return working_dir


@pytest.fixture(name="flowcell_runs_working_directory_dragen")
def fixture_flowcell_runs_working_directory_dragen(flowcell_runs_working_directory: Path):
    """Return the path to a working directory with flowcells ready for demux"""
    working_dir = flowcell_runs_working_directory / "dragen"
    working_dir.mkdir(parents=True)
    return working_dir


@pytest.fixture(name="demultiplexed_flowcells_working_directory")
def fixture_demultiplexed_flowcells_working_directory(project_dir: Path) -> Path:
    """Return the path to a working directory with flowcells that have been demultiplexed"""
    working_dir = project_dir / "demultiplexed-runs"
    working_dir.mkdir(parents=True)
    return working_dir


@pytest.fixture(name="demultiplexed_flowcell_working_directory")
def fixture_demultiplexed_flowcell_working_directory(
    demux_results_not_finished_dir: Path,
    demultiplexed_flowcells_working_directory: Path,
    flowcell_full_name: str,
) -> Path:
    """Copy the content of a demultiplexed but not finished directory to a temporary location"""
    source: Path = demux_results_not_finished_dir / flowcell_full_name
    destination: Path = demultiplexed_flowcells_working_directory / flowcell_full_name
    shutil.copytree(src=source, dst=destination)
    return destination


@pytest.fixture(name="demultiplexed_flowcell_finished_working_directory")
def fixture_demultiplexed_flowcell_finished_working_directory(
    demux_results_finished_dir: Path,
    demultiplexed_flowcells_working_directory: Path,
    flowcell_full_name: str,
) -> Path:
    """Copy the content of a demultiplexed but not finished directory to a temporary location"""
    source: Path = demux_results_finished_dir / flowcell_full_name
    destination: Path = demultiplexed_flowcells_working_directory / flowcell_full_name
    shutil.copytree(src=source, dst=destination)
    return destination


@pytest.fixture(name="flowcell_working_directory")
def fixture_flowcell_working_directory(
    novaseq_dir: Path, flowcell_runs_working_directory: Path
) -> Path:
    """Return the path to a working directory that will be deleted after test is run

    This is a path to a flowcell directory with the run parameters present
    """
    working_dir: Path = flowcell_runs_working_directory / novaseq_dir.name
    working_dir.mkdir(parents=True)
    existing_flowcell: Flowcell = Flowcell(flowcell_path=novaseq_dir)
    working_flowcell: Flowcell = Flowcell(flowcell_path=working_dir)
    shutil.copy(
        str(existing_flowcell.run_parameters_path), str(working_flowcell.run_parameters_path)
    )
    return working_dir


@pytest.fixture(name="flowcell_working_directory_bcl2fastq")
def fixture_flowcell_working_directory_bcl2fastq(
    flowcell_dir_bcl2fastq: Path, flowcell_runs_working_directory_bcl2fastq: Path
) -> Path:
    """Return the path to a working directory that will be deleted after test is run

    This is a path to a flowcell directory with the run parameters present
    """
    working_dir: Path = flowcell_runs_working_directory_bcl2fastq / flowcell_dir_bcl2fastq.name
    working_dir.mkdir(parents=True)
    existing_flowcell: Flowcell = Flowcell(flowcell_path=flowcell_dir_bcl2fastq)
    working_flowcell: Flowcell = Flowcell(flowcell_path=working_dir)
    shutil.copy(
        str(existing_flowcell.run_parameters_path), str(working_flowcell.run_parameters_path)
    )
    return working_dir


@pytest.fixture(name="flowcell_working_directory_dragen")
def fixture_flowcell_working_directory_dragen(
    flowcell_dir_dragen: Path, flowcell_runs_working_directory_dragen: Path
) -> Path:
    """Return the path to a working directory that will be deleted after test is run

    This is a path to a flowcell directory with the run parameters present
    """
    working_dir: Path = flowcell_runs_working_directory_dragen / flowcell_dir_dragen.name
    working_dir.mkdir(parents=True)
    existing_flowcell: Flowcell = Flowcell(flowcell_path=flowcell_dir_dragen)
    working_flowcell: Flowcell = Flowcell(flowcell_path=working_dir)
    shutil.copy(
        str(existing_flowcell.run_parameters_path), str(working_flowcell.run_parameters_path)
    )
    return working_dir


@pytest.fixture(name="flowcell_working_directory_no_run_parameters")
def fixture_flowcell_working_directory_no_run_parameters(
    novaseq_dir: Path, flowcell_runs_working_directory: Path
) -> Path:
    """This is a path to a flowcell directory with the run parameters missing"""
    working_dir: Path = flowcell_runs_working_directory / novaseq_dir.name
    working_dir.mkdir(parents=True)

    return working_dir


@pytest.fixture(name="demultiplex_ready_flowcell")
def fixture_demultiplex_ready_flowcell(flowcell_working_directory: Path, novaseq_dir: Path) -> Path:
    """Return the path to a working directory that is ready for demultiplexing

    This is a path to a flowcell directory with all the files necessary to start demultiplexing present
    """
    existing_flowcell: Flowcell = Flowcell(flowcell_path=novaseq_dir)
    working_flowcell: Flowcell = Flowcell(flowcell_path=flowcell_working_directory)
    shutil.copy(str(existing_flowcell.sample_sheet_path), str(working_flowcell.sample_sheet_path))
    shutil.copy(
        str(DemultiplexingAPI.get_stderr_logfile(existing_flowcell)),
        str(DemultiplexingAPI.get_stderr_logfile(working_flowcell)),
    )
    working_flowcell.copy_complete_path.touch()
    working_flowcell.rta_complete_path.touch()
    return flowcell_working_directory


@pytest.fixture(name="demultiplex_ready_flowcell_bcl2fastq")
def fixture_demultiplex_ready_flowcell_bcl2fastq(
    flowcell_working_directory_bcl2fastq: Path, flowcell_dir_bcl2fastq: Path
) -> Path:
    """Return the path to a working directory that is ready for demultiplexing

    This is a path to a flowcell directory with all the files necessary to start demultiplexing present
    """
    existing_flowcell: Flowcell = Flowcell(flowcell_path=flowcell_dir_bcl2fastq)
    working_flowcell: Flowcell = Flowcell(flowcell_path=flowcell_working_directory_bcl2fastq)
    shutil.copy(str(existing_flowcell.sample_sheet_path), str(working_flowcell.sample_sheet_path))
    shutil.copy(
        str(DemultiplexingAPI.get_stderr_logfile(existing_flowcell)),
        str(DemultiplexingAPI.get_stderr_logfile(working_flowcell)),
    )
    working_flowcell.copy_complete_path.touch()
    working_flowcell.rta_complete_path.touch()
    return flowcell_working_directory_bcl2fastq


@pytest.fixture(name="demultiplex_ready_flowcell_dragen")
def fixture_demultiplex_ready_flowcell_dragen(
    flowcell_working_directory_dragen: Path, flowcell_dir_dragen: Path
) -> Path:
    """Return the path to a working directory that is ready for demultiplexing

    This is a path to a flowcell directory with all the files necessary to start demultiplexing present
    """
    existing_flowcell: Flowcell = Flowcell(
        flowcell_path=flowcell_dir_dragen, bcl_converter="dragen"
    )
    working_flowcell: Flowcell = Flowcell(
        flowcell_path=flowcell_working_directory_dragen, bcl_converter="dragen"
    )
    shutil.copy(str(existing_flowcell.sample_sheet_path), str(working_flowcell.sample_sheet_path))
    shutil.copy(
        str(DemultiplexingAPI.get_stderr_logfile(existing_flowcell)),
        str(DemultiplexingAPI.get_stderr_logfile(working_flowcell)),
    )
    working_flowcell.copy_complete_path.touch()
    working_flowcell.rta_complete_path.touch()
    return flowcell_working_directory_dragen


@pytest.fixture(name="sample_sheet_context")
def fixture_sample_sheet_context(cg_context: CGConfig, lims_api: LimsAPI) -> CGConfig:
    cg_context.lims_api_ = lims_api
    return cg_context


@pytest.fixture(name="demultiplex_configs")
def fixture_demultiplex_configs(
    flowcell_runs_working_directory: Path,
    demultiplexed_flowcells_working_directory: Path,
    demultiplex_fixtures: Path,
) -> dict:
    out_dir: Path = demultiplexed_flowcells_working_directory
    run_dir: Path = flowcell_runs_working_directory
    out_dir.mkdir(parents=True, exist_ok=True)
    return {
        "demultiplex": {
            "out_dir": str(out_dir),
            "run_dir": str(run_dir),
            "slurm": {"account": "test", "mail_user": "testuser@github.se"},
        }
    }


@pytest.fixture(name="demultiplexing_api")
def fixture_demultiplexing_api(
    demultiplex_configs: dict, sbatch_process: Process
) -> DemultiplexingAPI:
    demux_api = DemultiplexingAPI(config=demultiplex_configs)
    demux_api.slurm_api.process = sbatch_process
    return demux_api


@pytest.fixture(name="demultiplex_context")
def fixture_demultiplex_context(
    demultiplexing_api: DemultiplexingAPI, stats_api: StatsAPI, cg_context: CGConfig
) -> CGConfig:
    cg_context.demultiplex_api_ = demultiplexing_api
    cg_context.cg_stats_api_ = stats_api
    return cg_context


@pytest.fixture(name="cli_runner")
def fixture_cli_runner():
    """Create a CliRunner"""
    return CliRunner()



"""Constants for delivery"""

import copy

ONLY_ONE_CASE_PER_TICKET = [
    "fastq",
    "microsalt",
    "sarscov2",
]

SKIP_MISSING = [
    "fastq",
    "microsalt",
    "sarscov2",
]

BALSAMIC_ANALYSIS_ONLY_CASE_TAGS = [
    {"cnvkit", "filtered", "sv-vcf"},
    {"cnvkit", "filtered", "sv-vcf-index"},
    {"cnvkit", "regions"},
    {"vcf-snv-clinical"},
    {"vcf-snv-clinical-index"},
    {"vardict", "deliver"},
    {"vcf", "sention", "haplotype-caller", "filtered"},
    {"vcf-index", "sention", "haplotype-caller", "filtered"},
    {"vcf-sv-clinical", "manta", "filtered"},
    {"vcf-sv-clinical-index", "manta", "filtered"},
    {"cnvkit", "visualization"},
    {"cnvkit", "visualization", "diagram"},
    {"multiqc-html"},
]

BALSAMIC_ANALYSIS_CASE_TAGS = copy.deepcopy(BALSAMIC_ANALYSIS_ONLY_CASE_TAGS)
BALSAMIC_ANALYSIS_CASE_TAGS.extend(
    [
        {"cram", "normal"},
        {"cram-index"},
        {"cram", "tumor"},
        {"cram-index", "tumor"},
    ]
)

BALSAMIC_ANALYSIS_SAMPLE_TAGS = [
    {"cram", "normal"},
    {"cram-index"},
    {"cram", "tumor"},
    {"cram-index", "tumor"},
]

BALSAMIC_QC_CASE_TAGS = [
    {"multiqc-html"},
]
BALSAMIC_QC_SAMPLE_TAGS = [
    {"fastq"},
]


MIP_DNA_ANALYSIS_CASE_TAGS = [
    {"vcf-clinical-sv-bin"},
    {"vcf-clinical-sv-bin-index"},
    {"vcf-research-sv-bin"},
    {"vcf-research-sv-bin-index"},
    {"gbcf"},
    {"gbcf-index"},
    {"snv-gbcf"},
    {"snv-gbcf-index"},
    {"snv-bcf"},
    {"snv-bcf-index"},
    {"sv-bcf"},
    {"sv-bcf-index"},
    {"vcf-snv-clinical"},
    {"vcf-snv-clinical-index"},
    {"vcf-snv-research"},
    {"vcf-snv-research-index"},
    {"vcf-sv-clinical"},
    {"vcf-sv-clinical-index"},
    {"vcf-sv-research"},
    {"vcf-sv-research-index"},
]

MIP_DNA_ANALYSIS_SAMPLE_TAGS = [{"bam"}, {"bam-index"}, {"cram"}, {"cram-index"}]

MIP_RNA_ANALYSIS_CASE_TAGS = [
    {"fusion", "clinical", "pdf"},
    {"fusion", "research", "pdf"},
    {"fusion", "vcf"},
    {"fusion", "vcf-index"},
    {"vcf-snv-clinical"},
    {"vcf-snv-clinical-index"},
    {"vcf-snv-research"},
    {"vcf-snv-research-index"},
]

MIP_RNA_ANALYSIS_SAMPLE_TAGS = [
    {"cram"},
    {"cram-index"},
    {"fusion", "vcf"},
    {"fusion", "vcf-index"},
    {"salmon-quant"},
]

MICROSALT_ANALYSIS_CASE_TAGS = [
    {"microsalt-qc"},
    {"microsalt-type"},
    {"assembly"},
    {"trimmed-forward-reads"},
    {"trimmed-reverse-reads"},
    {"reference-alignment-deduplicated"},
]

MICROSALT_ANALYSIS_SAMPLE_TAGS = []

FASTQ_ANALYSIS_CASE_TAGS = []

FASTQ_ANALYSIS_SAMPLE_TAGS = [
    {"fastq"},
]

SARSCOV2_ANALYSIS_CASE_TAGS = [
    {"pangolin"},
    {"ks-delivery"},
]

SARSCOV2_ANALYSIS_SAMPLE_TAGS = [
    {"fastq"},
]

PIPELINE_ANALYSIS_TAG_MAP = {
    "balsamic": {
        "case_tags": BALSAMIC_ANALYSIS_CASE_TAGS,
        "sample_tags": BALSAMIC_ANALYSIS_SAMPLE_TAGS,
    },
    "balsamic-analysis": {"case_tags": BALSAMIC_ANALYSIS_ONLY_CASE_TAGS, "sample_tags": []},
    "balsamic-qc": {
        "case_tags": BALSAMIC_QC_CASE_TAGS,
        "sample_tags": BALSAMIC_QC_SAMPLE_TAGS,
    },
    "mip-dna": {
        "case_tags": MIP_DNA_ANALYSIS_CASE_TAGS,
        "sample_tags": MIP_DNA_ANALYSIS_SAMPLE_TAGS,
    },
    "mip-rna": {
        "case_tags": MIP_RNA_ANALYSIS_CASE_TAGS,
        "sample_tags": MIP_RNA_ANALYSIS_SAMPLE_TAGS,
    },
    "microsalt": {
        "case_tags": MICROSALT_ANALYSIS_CASE_TAGS,
        "sample_tags": MICROSALT_ANALYSIS_SAMPLE_TAGS,
    },
    "fastq": {
        "case_tags": FASTQ_ANALYSIS_CASE_TAGS,
        "sample_tags": FASTQ_ANALYSIS_SAMPLE_TAGS,
    },
    "sarscov2": {
        "case_tags": SARSCOV2_ANALYSIS_CASE_TAGS,
        "sample_tags": SARSCOV2_ANALYSIS_SAMPLE_TAGS,
    },
}

PIPELINE_ANALYSIS_OPTIONS = PIPELINE_ANALYSIS_TAG_MAP.keys()



"""Fixtures for utils"""

import pytest

from cg.utils import Process


@pytest.fixture(scope="function")
def ls_process():
    """
    list files process
    """
    binary = "ls"
    process = Process(binary=binary)
    return process


@pytest.fixture(scope="function")
def echo_process():
    """
    echo process
    """
    binary = "echo"
    process = Process(binary=binary)
    return process


@pytest.fixture(scope="function")
def stderr_output():
    """
    std err lines
    """
    lines = (
        "2018-11-29 08:41:38 130-229-8-20-dhcp.local "
        "mongo_adapter.client[77135] INFO Connecting to "
        "uri:mongodb://None:None@localhost:27017\n"
        "2018-11-29 08:41:38 130-229-8-20-dhcp.local "
        "mongo_adapter.client[77135] INFO Connection "
        "established\n"
    )
    return lines



import logging
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import openpyxl
from openpyxl.cell.cell import Cell
from openpyxl.workbook import Workbook
from openpyxl.worksheet.worksheet import Worksheet
from pydantic import parse_obj_as

from cg.apps.orderform.orderform_parser import OrderformParser
from cg.constants import DataDelivery
from cg.exc import OrderFormError
from cg.meta.orders import OrderType
from cg.models.orders.excel_sample import ExcelSample

LOG = logging.getLogger(__name__)


class ExcelOrderformParser(OrderformParser):
    NO_ANALYSIS: str = "no-analysis"
    NO_VALUE: str = "no_value"
    SHEET_NAMES: List[str] = ["Orderform", "orderform", "order form"]
    VALID_ORDERFORMS: List[str] = [
        "1508:25",  # Orderform MIP, Balsamic, sequencing only, MIP RNA
        "1603:10",  # Microbial WGS
        "1604:10",  # Orderform Ready made libraries (RML)
        "1605:9",  # Microbial meta genomes
        "2184:5",  # Orderform SARS-CoV-2
    ]

    def check_orderform_version(self, document_title: str) -> None:
        """Raise an error if the orderform is too new or too old for the order portal."""
        LOG.info("Validating that %s is a correct orderform version", document_title)
        for valid_orderform in self.VALID_ORDERFORMS:
            if valid_orderform in document_title:
                return
        raise OrderFormError(f"Unsupported orderform: {document_title}")

    def get_sheet_name(self, sheet_names: List[str]) -> str:
        """Return the correct (existing) sheet names"""

        for name in sheet_names:
            if name in self.SHEET_NAMES:
                LOG.info("Found sheet name %s", name)
                return name
        raise OrderFormError("'orderform' sheet not found in Excel file")

    @staticmethod
    def get_document_title(workbook: Workbook, orderform_sheet: Worksheet) -> str:
        """Get the document title for the order form.

        Openpyxl use 1 based counting
        """
        for sheet_number, sheet_name in enumerate(workbook.sheetnames):
            if sheet_name.lower() != "information":
                continue
            information_sheet: Worksheet = workbook[sheet_name]
            document_title = information_sheet.cell(1, 3).value
            LOG.info("Found document title %s", document_title)
            return document_title

        document_title = orderform_sheet.cell(1, 2).value
        LOG.info("Found document title %s", document_title)
        return document_title

    @staticmethod
    def get_sample_row_info(
        row: Tuple[Cell], header_row: List[str], empty_row_found: bool
    ) -> Optional[dict]:
        """Convert an excel row with sample data into a dict with sample info"""
        values = []
        cell: Cell
        for cell in row:
            value = str(cell.value)
            if value == "None":
                value = ""
            if value == "NA":
                value = None
            values.append(value)

        if not values[0]:
            return None
        # skip empty rows
        if empty_row_found:
            raise OrderFormError(
                f"Found data after empty lines. Please delete any "
                f"non-sample data rows in between the samples"
            )

        sample_dict = dict(zip(header_row, values))
        sample_dict.pop(None)
        return sample_dict

    @staticmethod
    def get_header(rows: List[Tuple[Cell]]) -> List[str]:
        header_row: List[str] = []
        header = False
        for row in rows:
            if header:
                return [cell.value for cell in row]
            if row[0].value == "<TABLE HEADER>":
                LOG.debug("Found header row")
                header = True
        return header_row

    @staticmethod
    def get_raw_samples(rows: List[Tuple[Cell]], header_row: List[str]) -> List[dict]:
        raw_samples: List[dict] = []
        sample_rows = False
        empty_row_found = False
        for row in rows:
            if row[0].value == "</SAMPLE ENTRIES>":
                LOG.debug("End of samples info")
                return raw_samples

            if sample_rows:
                sample_dict: Optional[dict] = ExcelOrderformParser.get_sample_row_info(
                    row=row, header_row=header_row, empty_row_found=empty_row_found
                )
                if sample_dict:
                    raw_samples.append(sample_dict)
                else:
                    empty_row_found = True

            elif row[0].value == "<SAMPLE ENTRIES>":
                LOG.debug("Found samples row")
                sample_rows = True
        return raw_samples

    @staticmethod
    def relevant_rows(orderform_sheet: Worksheet) -> List[Dict[str, str]]:
        """Get the relevant rows from an order form sheet."""
        # orderform_sheet.rows is a generator. Convert to list to be able to iterate multiple times
        rows = list(orderform_sheet.rows)
        header_row: List[str] = ExcelOrderformParser.get_header(rows)
        return ExcelOrderformParser.get_raw_samples(rows=rows, header_row=header_row)

    def get_project_type_from_samples(self) -> str:
        analyses = {sample.data_analysis for sample in self.samples}

        if len(analyses) != 1:
            raise OrderFormError(f"mixed 'Data Analysis' types: {', '.join(analyses)}")

        analysis = analyses.pop().lower().replace(" ", "-")
        return "fastq" if analysis == self.NO_ANALYSIS else analysis

    def get_project_type(self, document_title: str) -> str:
        """Determine the project type and set it to the class."""
        document_number_to_project_type = {
            "1604": str(OrderType.RML),
            "1603": str(OrderType.MICROSALT),
            "1605": str(OrderType.METAGENOME),
            "2184": str(OrderType.SARS_COV_2),
        }
        for document_number, value in document_number_to_project_type.items():
            if document_number in document_title:
                return value

        if "1508" in document_title:
            return self.get_project_type_from_samples()

        raise OrderFormError(f"Undetermined project type in: {document_title}")

    def parse_data_analysis(self) -> str:
        data_analyses = {sample.data_analysis for sample in self.samples if sample.data_analysis}

        if len(data_analyses) > 1:
            raise OrderFormError(f"mixed 'Data Analysis' types: {', '.join(data_analyses)}")

        return data_analyses.pop().lower().replace(" ", "-")

    def is_from_orderform_without_data_delivery(self, data_delivery: str) -> bool:
        return data_delivery == self.NO_VALUE

    def default_delivery_type(self, project_type: OrderType) -> str:
        """Returns the default delivery type for a project type"""
        if project_type == OrderType.FLUFFY:
            return DataDelivery.NIPT_VIEWER

        if project_type == OrderType.METAGENOME:
            return DataDelivery.FASTQ

        if project_type == OrderType.MICROSALT:
            data_analysis: str = self.parse_data_analysis()

            if data_analysis == "custom":
                return DataDelivery.FASTQ_QC

            if data_analysis == "fastq":
                return DataDelivery.FASTQ_QC

        if project_type == OrderType.RML:
            return DataDelivery.FASTQ

        raise OrderFormError(f"Could not determine value for Data Delivery")

    def get_data_delivery(self, project_type: OrderType) -> str:
        """Determine the order_data delivery type."""

        data_delivery: str = self.parse_data_delivery()

        if self.is_from_orderform_without_data_delivery(data_delivery):
            return self.default_delivery_type(project_type)

        if data_delivery == "fastq-qc":
            return DataDelivery.FASTQ_QC
        if data_delivery == "fastq-qc-+-analysis-+-cram":
            return DataDelivery.FASTQ_QC_ANALYSIS_CRAM
        if data_delivery == "fastq-qc-+-analysis-+-cram-+-scout":
            return DataDelivery.FASTQ_QC_ANALYSIS_CRAM_SCOUT

        try:
            return DataDelivery(data_delivery)
        except ValueError:
            raise OrderFormError(f"Unsupported Data Delivery: {data_delivery}")

    def parse_data_delivery(self) -> str:

        data_deliveries = {sample.data_delivery or self.NO_VALUE for sample in self.samples}

        if len(data_deliveries) > 1:
            raise OrderFormError(f"mixed 'Data Delivery' types: {', '.join(data_deliveries)}")

        return data_deliveries.pop().lower().replace(" ", "-")

    def get_customer_id(self) -> str:
        """Set the customer id"""

        customers = {sample.customer for sample in self.samples}
        if len(customers) != 1:
            raise OrderFormError("Invalid customer information: {}".format(customers))
        return customers.pop()

    def parse_orderform(self, excel_path: str) -> None:
        """Parse out information from an order form."""

        LOG.info("Open excel workbook from file %s", excel_path)
        workbook: Workbook = openpyxl.load_workbook(
            filename=excel_path, read_only=True, data_only=True
        )

        sheet_name: str = self.get_sheet_name(workbook.sheetnames)

        orderform_sheet: Worksheet = workbook[sheet_name]
        document_title: str = self.get_document_title(
            workbook=workbook, orderform_sheet=orderform_sheet
        )
        self.check_orderform_version(document_title)

        LOG.info("Parsing all samples from orderform")
        raw_samples: List[dict] = self.relevant_rows(orderform_sheet)

        if not raw_samples:
            raise OrderFormError("orderform doesn't contain any samples")

        self.samples: List[ExcelSample] = parse_obj_as(List[ExcelSample], raw_samples)
        self.project_type: str = self.get_project_type(document_title)
        self.delivery_type = self.get_data_delivery(project_type=OrderType(self.project_type))
        self.customer_id = self.get_customer_id()

        self.order_name = Path(excel_path).stem



"""Tests for cleaning FASTQ files"""
import logging

import pytest

from cg.meta.compress import files


@pytest.mark.compress_meta
def test_remove_fastqs(compress_api, compression_object, caplog):
    """Test remove_fastq method

    This method should remove FASTQ files since compression is completed
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN existing FASTQ and flag file
    fastq_first = compression_object.fastq_first
    fastq_second = compression_object.fastq_second
    fastq_first.touch()
    fastq_second.touch()
    compression_object.spring_metadata_path.touch()

    # WHEN calling remove_fastq
    compress_api.remove_fastq(fastq_first=fastq_first, fastq_second=fastq_second)

    # THEN the assert that the FASTQ-files are deleted
    assert not fastq_first.exists()
    assert not fastq_second.exists()
    # THEN assert that the flag file is still there since this holds important information
    assert compression_object.spring_metadata_path.exists()
    expected_output = f"Will remove {fastq_first} and {fastq_second}"
    assert expected_output in caplog.text
    assert "FASTQ files removed" in caplog.text


@pytest.mark.compress_meta
def test_update_hk_fastq(real_housekeeper_api, compress_hk_fastq_bundle, compress_api, helpers):
    """Test to update the FASTQ and SPRING paths in housekeeper after completed compression

    This will test so that the FASTQ files are replaced by a SPRING file and a SPRING metadata file
    """
    # GIVEN real housekeeper api populated with a housekeeper bundle
    hk_bundle = compress_hk_fastq_bundle
    sample_id = hk_bundle["name"]
    hk_api = real_housekeeper_api
    helpers.ensure_hk_bundle(real_housekeeper_api, hk_bundle)
    compress_api.hk_api = hk_api

    # GIVEN that there are some FASTQ files in housekeeper
    hk_fastq_files = list(hk_api.files(tags=["fastq"]))
    assert hk_fastq_files
    # GIVEN that there are no SPRING files in housekeeper
    hk_spring_file = list(hk_api.files(tags=["spring"]))
    assert not hk_spring_file
    hk_fastq_flag_file = list(hk_api.files(tags=["spring-metadata"]))
    assert not hk_fastq_flag_file
    # GIVEN a housekeeper version
    version_obj = compress_api.get_latest_version(sample_id)
    fastq_dict = files.get_fastq_files(sample_id=sample_id, version_obj=version_obj)
    run = list(fastq_dict.keys())[0]
    compression_obj = fastq_dict[run]["compression_data"]

    # WHEN updating hk
    compress_api.update_fastq_hk(
        sample_id=sample_id,
        compression_obj=compression_obj,
        hk_fastq_first=fastq_dict[run]["hk_first"],
        hk_fastq_second=fastq_dict[run]["hk_second"],
    )

    # THEN assert that the FASTQ files are removed from housekeeper
    hk_fastq_files = list(hk_api.files(tags=["fastq"]))
    assert not hk_fastq_files
    # THEN assert that the SPRING file and the metadata file is added to housekeeper
    hk_spring_file = list(real_housekeeper_api.files(tags=["spring"]))
    assert hk_spring_file
    hk_fastq_flag_file = list(real_housekeeper_api.files(tags=["spring-metadata"]))
    assert hk_fastq_flag_file


@pytest.mark.compress_meta
@pytest.mark.clean_fastq
def test_cli_clean_fastqs_removed(populated_compress_fastq_api, compression_files, sample):
    """Test to clean FASTQs after a succesfull FASTQ compression

    Files should be cleaned since the compression is completed
    """
    compress_api = populated_compress_fastq_api
    spring_file = compression_files.spring_file
    spring_metadata_file = compression_files.spring_metadata_file
    fastq_first = compression_files.fastq_first_file
    fastq_second = compression_files.fastq_second_file
    # GIVEN that the SPRING compression files exist
    assert spring_file.exists()
    assert spring_metadata_file.exists()
    # GIVEN that the FASTQ files exists
    assert fastq_first.exists()
    assert fastq_second.exists()

    # WHEN running the clean command
    compress_api.clean_fastq(sample)

    # THEN assert SPRING files exists
    assert spring_file.exists()
    assert spring_metadata_file.exists()
    # THEN assert that the FASTQ files are removed
    assert not fastq_first.exists()
    assert not fastq_second.exists()


@pytest.mark.compress_meta
@pytest.mark.clean_fastq
def test_cli_clean_fastqs_no_spring_metadata(
    populated_compress_fastq_api, sample, compression_files
):
    """Test to clean FASTQs when SPRING compression is not finished

    No files should be cleaned since compression is not completed
    """
    compress_api = populated_compress_fastq_api
    spring_file = compression_files.spring_file
    spring_metadata_file = compression_files.spring_metadata_path
    fastq_first = compression_files.fastq_first_file
    fastq_second = compression_files.fastq_second_file
    # GIVEN that the SPRING compression file exist
    assert spring_file.exists()
    # GIVEN that the SPRING metadata file does not exist
    assert not spring_metadata_file.exists()

    # WHEN running the clean command
    compress_api.clean_fastq(sample)

    # THEN assert SPRING file exists
    assert spring_file.exists()
    # THEN assert that the FASTQ files are NOT removed
    assert fastq_first.exists()
    assert fastq_second.exists()


@pytest.mark.compress_meta
@pytest.mark.clean_fastq
def test_cli_clean_fastqs_pending_compression_metadata(
    populated_compress_fastq_api, sample, compression_files
):
    """Test to clean FASTQs when SPRING compression is pending

    No files should be cleaned since compression is not completed
    """
    compress_api = populated_compress_fastq_api
    spring_file = compression_files.spring_file
    fastq_first = compression_files.fastq_first_file
    fastq_second = compression_files.fastq_second_file

    # GIVEN that the SPRING compression file exist
    assert spring_file.exists()
    crunchy_flag_file = compression_files.pending_file
    assert crunchy_flag_file.exists()

    # WHEN running the clean command
    compress_api.clean_fastq(sample)

    # THEN assert SPRING file exists
    assert spring_file.exists()
    # THEN assert that the FASTQ files are NOT removed
    assert fastq_first.exists()
    assert fastq_second.exists()



from pathlib import Path

import pytest

from cg.models.observations.observations_input_files import (
    ObservationsInputFiles,
    check_observation_file_from_hk,
)


def test_instantiate_input_files(observation_input_files_raw: dict):
    """Tests input files against a pydantic ObservationsInputFiles"""
    # GIVEN a dictionary with the basic config

    # WHEN instantiating a MipBaseConfig object
    files_object = ObservationsInputFiles(**observation_input_files_raw)

    # THEN assert that it was successfully created
    assert isinstance(files_object, ObservationsInputFiles)


def test_input_files_case_id(case_id: str, observation_input_files_raw: dict):
    """Test case_id"""
    # GIVEN a dictionary with the basic input files

    # WHEN instantiating a ObservationsInputFiles object
    files_object = ObservationsInputFiles(**observation_input_files_raw)

    # THEN assert that case_id was set
    assert files_object.case_id == case_id


def test_check_observation_file_from_hk(filled_file: Path, hk_tag: str):
    """Test check_observation_file_from_hk"""
    # GIVEN a file path which exists

    # GIVEN a Housekeeper tag

    # WHEN checking the observation file
    is_ok = check_observation_file_from_hk(file_tag=hk_tag, file=filled_file)

    # THEN if file is ok return true
    assert is_ok


def test_check_observation_file_from_hk_when_no_hk_file_exists(hk_tag: str):
    """Test check_observation_file_from_hk when file does not exist"""
    # GIVEN a file path which is None
    file = None

    # GIVEN a Housekeeper tag

    with pytest.raises(FileNotFoundError):
        # WHEN checking the observation file

        # THEN FileNotFoundError should be raised
        check_observation_file_from_hk(file_tag=hk_tag, file=file)


def test_check_observation_file_from_hk_when_file_not_exists(
    file_does_not_exist: Path, hk_tag: str
):
    """Test check_observation_file_from_hk when file does not exist"""
    # GIVEN a file path which does not exists

    # GIVEN a Housekeeper tag

    # THEN check that file does not exit
    assert file_does_not_exist.exists() is False

    with pytest.raises(FileNotFoundError):
        # WHEN checking the observation file

        # THEN FileNotFoundError should be raised
        check_observation_file_from_hk(file_tag=hk_tag, file=file_does_not_exist)


def test_input_files_pedigree(filled_file: Path, observation_input_files_raw: dict):
    """Test pedigree validator"""
    # GIVEN a dictionary with the basic input files

    # WHEN instantiating a ObservationsInputFiles object
    files_object = ObservationsInputFiles(**observation_input_files_raw)

    # THEN assert that pedigree file was set
    assert files_object.pedigree == filled_file


def test_input_files_pedigree_not_exists(
    file_does_not_exist: Path, observation_input_files_raw: dict
):
    """Test pedigree validator when file does not exits"""
    # GIVEN a dictionary with the basic input files

    # WHEN pedigree file does not exist in file system
    observation_input_files_raw["pedigree"] = file_does_not_exist

    assert Path(observation_input_files_raw["pedigree"]).exists() is False

    with pytest.raises(FileNotFoundError):
        # WHEN instantiating a ObservationsInputFiles object
        files_object = ObservationsInputFiles(**observation_input_files_raw)

        assert files_object.pedigree.exists() is False


def test_input_files_sv_vcf_when_none(observation_input_files_raw: dict):
    """Test sv_vcf validator when file is None"""
    # GIVEN a dictionary with the basic input files

    # WHEN instantiating a ObservationsInputFiles object
    files_object = ObservationsInputFiles(**observation_input_files_raw)

    assert files_object.sv_vcf is None


def test_input_files_sv_vcf_when_file(filled_file: Path, observation_input_files_raw: dict):
    """Test sv_vcf validator when file is supplied and exists"""
    # GIVEN a dictionary with the basic input files

    # GIVEN a sv_vcf file path
    observation_input_files_raw["sv_vcf"] = filled_file

    # WHEN instantiating a ObservationsInputFiles object
    files_object = ObservationsInputFiles(**observation_input_files_raw)

    assert files_object.sv_vcf == filled_file



from cg.cli.add import add
from cg.models.cg_config import CGConfig
from cg.store import Store
from click.testing import CliRunner


def test_add_customer(cli_runner: CliRunner, base_context: CGConfig):
    # GIVEN database with some customers
    status_db: Store = base_context.status_db
    nr_customers: int = status_db.Customer.query.count()

    # WHEN adding a customer
    cli_runner.invoke(
        add,
        [
            "customer",
            "internal_id",
            "testcust",
            "--invoice-address",
            "Test adress",
            "--invoice-reference",
            "ABCDEF",
        ],
        obj=base_context,
    )

    # THEN it should be stored in the database
    assert status_db.Customer.query.count() == nr_customers + 1


def test_add_user(cli_runner: CliRunner, base_context: CGConfig):
    # GIVEN a database with a customer in it that we can connect the user to
    disk_store: Store = base_context.status_db
    customer_id = "custtest"
    customer_group = disk_store.add_customer_group("dummy_group", "dummy group")
    customer = disk_store.add_customer(
        internal_id=customer_id,
        name="Test Customer",
        scout_access=False,
        customer_group=customer_group,
        invoice_address="Street nr, 12345 Uppsala",
        invoice_reference="ABCDEF",
    )
    disk_store.add_commit(customer)
    # GIVEN that there is a certain number of users
    nr_users = disk_store.User.query.count()

    # WHEN adding a new user
    name, email = "Paul T. Anderson", "paul.anderson@magnolia.com"
    result = cli_runner.invoke(add, ["user", "-c", customer_id, email, name], obj=base_context)

    # THEN it should be stored in the database
    assert result.exit_code == 0
    assert disk_store.User.query.count() == nr_users + 1



import datetime

import pytest
from pydantic import ValidationError

from cg.constants import DataDelivery
from cg.models.orders.excel_sample import ExcelSample


def test_excel_minimal_sample_schema(minimal_excel_sample: dict):
    """Test instantiate a minimal sample"""
    # GIVEN some simple sample info from the simplest sample possible

    # WHEN creating a excel sample
    excel_sample: ExcelSample = ExcelSample(**minimal_excel_sample)

    # THEN assert that the sample name info was correctly parsed
    assert excel_sample.name == minimal_excel_sample["Sample/Name"]


def test_excel_swedish_priority(minimal_excel_sample: dict):
    """Test instantiate a sample with swedish priority"""
    # GIVEN some sample where the priority is in swedish
    minimal_excel_sample["UDF/priority"] = "förtur"

    # WHEN creating a excel sample
    excel_sample: ExcelSample = ExcelSample(**minimal_excel_sample)

    # THEN assert that the sample priority info was correctly parsed
    assert excel_sample.priority == "priority"


def test_excel_source_type(minimal_excel_sample: dict):
    """Test instantiate a sample with source type"""
    # GIVEN some sample with a known source type
    source_type = "blood"
    minimal_excel_sample["UDF/Source"] = source_type

    # WHEN creating a excel sample
    excel_sample: ExcelSample = ExcelSample(**minimal_excel_sample)

    # THEN assert that the source type info was correctly parsed
    assert excel_sample.source == source_type


def test_excel_unknown_source_type(minimal_excel_sample: dict):
    """Test instantiate a sample with a unknown source type

    ValidationError should be raised since the source type does not exist
    """
    # GIVEN some sample with a known source type
    source_type = "flagalella"
    minimal_excel_sample["UDF/Source"] = source_type

    # WHEN creating a excel sample
    with pytest.raises(ValidationError):
        # THEN assert that a validation error is raised since source does not exist
        excel_sample: ExcelSample = ExcelSample(**minimal_excel_sample)


def test_excel_datetime_to_date_conversion(minimal_excel_sample: dict):
    """Test instantiate a sample with a datetime collection date

    Datetime string should be converted to date
    """
    # GIVEN some sample with a collection date as a datetime-string
    collection_date = "2021-05-05 00:00:00"
    minimal_excel_sample["UDF/Collection Date"] = collection_date

    # WHEN creating a excel sample
    excel_sample: ExcelSample = ExcelSample(**minimal_excel_sample)

    # THEN the collection_date should only contain the date
    assert str(excel_sample.collection_date) == "2021-05-05"


def test_excel_with_panels(minimal_excel_sample: dict):
    """Test instantiate a sample with some gene panels set"""
    # GIVEN some sample with two gene panels in a semi colon separated string
    panel_1 = "OMIM"
    panel_2 = "PID"
    minimal_excel_sample["UDF/Gene List"] = ";".join([panel_1, panel_2])

    # WHEN creating a excel sample
    excel_sample: ExcelSample = ExcelSample(**minimal_excel_sample)

    # THEN assert that the panels was parsed into a list
    assert set(excel_sample.panels) == set([panel_1, panel_2])


def test_mip_rna_sample_is_correct(mip_rna_orderform_sample: dict):
    """Test that a mip rna orderform sample in parsed correct

    This will test if the aliases are mapped to the correct field in the pydantic model
    """
    # GIVEN sample data about a known balsamic sample

    # WHEN parsing the sample
    mip_rna_sample: ExcelSample = ExcelSample(**mip_rna_orderform_sample)

    # THEN assert that the sample information is parsed correct
    assert mip_rna_sample.name == "s1"
    assert mip_rna_sample.container == "96 well plate"
    assert mip_rna_sample.data_analysis == "MIP RNA"
    assert mip_rna_sample.data_delivery.lower() == str(DataDelivery.ANALYSIS_FILES)
    assert mip_rna_sample.application == "RNAPOAR025"
    assert mip_rna_sample.sex == "male"
    # case-id on the case
    # customer on the order (data)
    # require-qc-ok on the case
    assert mip_rna_sample.source == "tissue (FFPE)"

    assert mip_rna_sample.container_name == "plate1"
    assert mip_rna_sample.well_position == "A:1"
    assert mip_rna_sample.tumour is True
    assert mip_rna_sample.quantity == "4"
    assert mip_rna_sample.comment == "other Elution buffer"
    assert mip_rna_sample.time_point == "0"



"""Module for deliver and rsync customer inbox on hasta to customer inbox on caesar"""
import datetime as dt
import logging
import os
from pathlib import Path
from typing import List, Tuple, Optional

from housekeeper.store.models import Bundle

from cg.apps.cgstats.db.models import Version
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.slurm.slurm_api import SlurmAPI
from cg.meta.meta import MetaAPI
from cg.meta.rsync.sbatch import RSYNC_COMMAND, ERROR_RSYNC_FUNCTION
from cg.models.cg_config import CGConfig
from cg.models.slurm.sbatch import Sbatch
from cg.store import models

LOG = logging.getLogger(__name__)


class ExternalDataAPI(MetaAPI):
    def __init__(self, config: CGConfig):
        super().__init__(config)
        self.hasta_path: str = config.external.hasta
        self.caesar_path: str = config.external.caesar
        self.base_path: str = config.data_delivery.base_path
        self.account: str = config.data_delivery.account
        self.mail_user: str = config.data_delivery.mail_user
        self.housekeeper_api: HousekeeperAPI = config.housekeeper_api

    def create_log_dir(self, ticket_id: int, dry_run: bool) -> Path:
        timestamp = dt.datetime.now()
        timestamp_str = timestamp.strftime("%y%m%d_%H_%M_%S_%f")
        folder_name = Path("_".join([str(ticket_id), timestamp_str]))
        log_dir = Path(self.base_path) / folder_name
        LOG.info("Creating folder: %s", log_dir)
        if dry_run:
            LOG.info("Would have created path %s, but this is a dry run", log_dir)
            return log_dir
        log_dir.mkdir(parents=True, exist_ok=False)
        return log_dir

    def create_source_path(
        self, cust_id: str, ticket_id: int, raw_path: str, cust_sample_id: str
    ) -> str:
        cust_id_added_to_path = (
            raw_path % cust_id + "/" + str(ticket_id) + "/" + cust_sample_id + "/"
        )
        return cust_id_added_to_path

    def create_destination_path(self, cust_id: str, raw_path: str, lims_sample_id: str) -> str:
        cust_id_added_to_path = raw_path % cust_id + "/" + lims_sample_id + "/"
        return cust_id_added_to_path

    def download_sample(
        self, cust_id: str, ticket_id: int, cust_sample_id: str, lims_sample_id: str, dry_run: bool
    ) -> int:
        log_dir: Path = self.create_log_dir(ticket_id=ticket_id, dry_run=dry_run)
        source_path: str = self.create_source_path(
            cust_id=cust_id,
            ticket_id=ticket_id,
            raw_path=self.caesar_path,
            cust_sample_id=cust_sample_id,
        )
        destination_path: str = self.create_destination_path(
            cust_id=cust_id,
            raw_path=self.hasta_path,
            lims_sample_id=lims_sample_id,
        )
        commands = RSYNC_COMMAND.format(source_path=source_path, destination_path=destination_path)
        error_function = ERROR_RSYNC_FUNCTION.format()
        sbatch_info = {
            "job_name": "_".join([str(ticket_id), "rsync_external_data"]),
            "account": self.account,
            "number_tasks": 1,
            "memory": 1,
            "log_dir": log_dir.as_posix(),
            "email": self.mail_user,
            "hours": 24,
            "commands": commands,
            "error": error_function,
        }
        slurm_api = SlurmAPI()
        slurm_api.set_dry_run(dry_run=dry_run)
        sbatch_content: str = slurm_api.generate_sbatch_content(Sbatch.parse_obj(sbatch_info))
        sbatch_path = log_dir / "_".join([str(ticket_id), "rsync_external_data.sh"])
        sbatch_number: int = slurm_api.submit_sbatch(
            sbatch_content=sbatch_content, sbatch_path=sbatch_path
        )
        return sbatch_number

    def download_ticket(self, ticket_id: int, dry_run: bool) -> None:
        cases: List[models.Family] = self.status_db.get_cases_from_ticket(ticket_id=ticket_id).all()
        cust_id = cases[0].customer.internal_id
        for case in cases:
            links = case.links
            for link in links:
                lims_sample_id = link.sample.internal_id
                cust_sample_id = link.sample.name
                sbatch_number = self.download_sample(
                    cust_id=cust_id,
                    ticket_id=ticket_id,
                    cust_sample_id=cust_sample_id,
                    lims_sample_id=lims_sample_id,
                    dry_run=dry_run,
                )
                LOG.info(
                    "Downloading sample %s from caesar by submitting sbatch job %s"
                    % (lims_sample_id, sbatch_number)
                )

    def get_all_fastq(self, sample_folder: str) -> List[str]:
        all_fastqs = []
        for leaf in os.listdir(sample_folder):
            abs_path = sample_folder + leaf
            LOG.info("Found file %s inside folder %s" % (abs_path, sample_folder))
            if abs_path.endswith("fastq.gz"):
                all_fastqs.append(abs_path)
        return all_fastqs

    def get_all_paths(self, lims_sample_id: str, cust_id: str) -> List[str]:
        fastq_folder: str = self.create_destination_path(
            cust_id=cust_id, lims_sample_id=lims_sample_id, raw_path=self.hasta_path
        )
        all_fastq_in_folder: List[str] = self.get_all_fastq(sample_folder=fastq_folder)
        return all_fastq_in_folder

    def create_hk_bundle(
        self, bundle_name: str, dry_run: bool, data_dict: dict
    ) -> Optional[Tuple[Bundle, Version]]:
        if dry_run:
            LOG.info(
                "Would have created bundle %s in housekeeper, but this is dry-run", bundle_name
            )
            return
        bundle_result: Tuple[Bundle, Version] = self.housekeeper_api.add_bundle(
            bundle_data=data_dict
        )
        return bundle_result

    def create_dict(self, name: str) -> dict:
        hk_dict = {
            "name": name,
            "created_at": dt.datetime.now(),
            "expires_at": None,
            "files": [],
        }
        return hk_dict

    def configure_housekeeper(self, ticket_id: int, dry_run: bool) -> None:
        cases = self.status_db.get_cases_from_ticket(ticket_id=ticket_id).all()
        cust_id = cases[0].customer.internal_id
        for case in cases:
            links = case.links
            for link in links:
                lims_sample_id = link.sample.internal_id
                paths: List[str] = self.get_all_paths(
                    lims_sample_id=lims_sample_id, cust_id=cust_id
                )
                if dry_run:
                    LOG.info(
                        "Would have added %s to housekeeper and linked associated files, but this is dry-run",
                        lims_sample_id,
                    )
                    return
                last_version = self.housekeeper_api.last_version(bundle=lims_sample_id)
                if not last_version:
                    LOG.info("Creating bundle for sample %s in housekeeper", lims_sample_id)
                    hk_dict: dict = self.create_dict(name=lims_sample_id)
                    bundle_result: Tuple[Bundle, Version] = self.create_hk_bundle(
                        bundle_name=lims_sample_id, data_dict=hk_dict, dry_run=dry_run
                    )
                    last_version = bundle_result[1]
                    self.housekeeper_api.add_commit(bundle_result[0], bundle_result[1])
                for path in paths:
                    LOG.info("Adding path %s to bundle %s in housekeeper" % (path, lims_sample_id))
                    self.housekeeper_api.add_file(
                        path=path, version_obj=last_version, tags=["fastq"]
                    )
                    self.housekeeper_api.commit()



import pytest
from cg.store import Store


@pytest.mark.parametrize("contact_type", ["delivery", "primary", "invoice"])
def test_contact_storing(store: Store, contact_type, helpers):

    # GIVEN an empty database
    assert store.Customer.query.first() is None
    internal_id, name, scout_access = "cust000", "Test customer", True
    customer_group = store.add_customer_group("dummy_group", "dummy group")
    contact_email = contact_type + ".contact@customer.se"
    contact_name = contact_type

    # WHEN adding a new customer and setting the new field
    new_customer = store.add_customer(
        internal_id=internal_id,
        name=name,
        scout_access=scout_access,
        customer_group=customer_group,
        invoice_address="dummy street 1",
        invoice_reference="dummy nr",
    )

    new_user = store.add_user(new_customer, contact_email, contact_name)

    contact_field = contact_type + "_contact"
    setattr(new_customer, contact_field, new_user)
    store.add_commit(new_customer)

    # THEN contact should be stored on the customer
    assert getattr(store.customer(internal_id=internal_id), contact_field) == new_user


@pytest.mark.parametrize("contact_type", ["delivery", "primary", "invoice"])
def test_contact_structure(store: Store, contact_type):

    # GIVEN the Customer model

    # WHEN checking the structure of Customer

    # THEN contact email and name should not be present
    assert not hasattr(store.Customer, contact_type + "_contact_email")
    assert not hasattr(store.Customer, contact_type + "_contact_name")

    # THEN deprecated attributes should not be present
    assert hasattr(store.Customer, contact_type + "_contact_id")
    assert hasattr(store.Customer, contact_type + "_contact")


def test_add_basic(store: Store):
    # GIVEN an empty database
    assert store.Customer.query.first() is None
    internal_id, name, scout_access = "cust000", "Test customer", True
    customer_group = store.add_customer_group("dummy_group", "dummy group")

    # WHEN adding a new customer
    new_customer = store.add_customer(
        internal_id=internal_id,
        name=name,
        scout_access=scout_access,
        customer_group=customer_group,
        invoice_address="dummy street 1",
        invoice_reference="dummy nr",
    )
    store.add_commit(new_customer)

    # THEN it should be stored in the database
    assert store.Customer.query.first() == new_customer



from pathlib import Path

import pytest

from cg.apps.cgstats.crud import create
from cg.apps.cgstats.stats import StatsAPI
from cg.models.demultiplex.demux_results import DemuxResults
from tests.apps.demultiplex.conftest import fixture_demultiplex_fixtures
from tests.models.demultiplexing.conftest import (
    fixture_bcl2fastq_demux_results,
    fixture_demultiplexed_flowcell,
    fixture_demultiplexed_runs,
    fixture_flowcell_full_name,
    fixture_flowcell_object,
    fixture_flowcell_path,
    fixture_flowcell_runs,
)


@pytest.fixture(name="stats_api")
def fixture_stats_api(project_dir: Path) -> StatsAPI:
    """Setup base CGStats store."""
    _store = StatsAPI({"cgstats": {"database": "sqlite://", "root": "tests/fixtures/DEMUX"}})
    _store.create_all()
    yield _store
    _store.drop_all()


@pytest.fixture(name="populated_stats_api")
def fixture_populated_stats_api(
    stats_api: StatsAPI, bcl2fastq_demux_results: DemuxResults
) -> StatsAPI:
    create.create_novaseq_flowcell(manager=stats_api, demux_results=bcl2fastq_demux_results)
    return stats_api


@pytest.fixture(name="demultiplexing_stats_path")
def fixture_demultiplexing_stats_path(bcl2fastq_demux_results: DemuxResults) -> Path:
    return bcl2fastq_demux_results.demux_stats_path


@pytest.fixture(name="conversion_stats_path")
def fixture_conversion_stats_path(bcl2fastq_demux_results: DemuxResults) -> Path:
    return bcl2fastq_demux_results.conversion_stats_path



from typing import List, Optional

from cg.constants import DEFAULT_CAPTURE_KIT, Pipeline
from cg.constants.gene_panel import GENOME_BUILD_37
from cg.meta.workflow.mip import MipAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store import models
from cg.utils import Process


class MipDNAAnalysisAPI(MipAnalysisAPI):
    def __init__(self, config: CGConfig, pipeline: Pipeline = Pipeline.MIP_DNA):
        super().__init__(config, pipeline)

    @property
    def root(self) -> str:
        return self.config.mip_rd_dna.root

    @property
    def conda_env(self) -> str:
        return self.config.mip_rd_dna.conda_env

    @property
    def mip_pipeline(self) -> str:
        return self.config.mip_rd_dna.pipeline

    @property
    def script(self) -> str:
        return self.config.mip_rd_dna.script

    @property
    def threshold_reads(self):
        return True

    @property
    def process(self) -> Process:
        if not self._process:
            self._process = Process(
                binary=f"{self.script} {self.mip_pipeline}",
                config=self.config.mip_rd_dna.mip_config,
                environment=self.conda_env,
            )
        return self._process

    def config_sample(self, link_obj: models.FamilySample, panel_bed: Optional[str]) -> dict:
        sample_data = self.get_sample_data(link_obj)
        if sample_data["analysis_type"] == "wgs":
            sample_data["capture_kit"] = panel_bed or DEFAULT_CAPTURE_KIT
        else:
            sample_data["capture_kit"] = panel_bed or self.get_target_bed_from_lims(
                link_obj.family.internal_id
            )
        if link_obj.mother:
            sample_data["mother"] = link_obj.mother.internal_id
        if link_obj.father:
            sample_data["father"] = link_obj.father.internal_id
        return sample_data

    def panel(self, case_id: str, genome_build: str = GENOME_BUILD_37) -> List[str]:
        """Create the aggregated gene panel file"""
        case_obj: models.Family = self.status_db.family(case_id)
        all_panels = self.convert_panels(case_obj.customer.internal_id, case_obj.panels)
        return self.scout_api.export_panels(build=genome_build, panels=all_panels)



"""Fixtures for the shipping api tests"""

from pathlib import Path

import pytest
from cg.apps.shipping import ShippingAPI
from cg.models.cg_config import CGConfig
from tests.apps.shipping.conftest import (
    fixture_binary_path,
    fixture_host_config,
    fixture_shipping_api,
    fixture_shipping_configs,
)


@pytest.fixture(name="fluffy_config")
def fixture_fluffy_config(apps_dir: Path) -> Path:
    """Return the path to a shipping config for fluffy"""
    return apps_dir / "shipping" / "fluffy-deploy.yaml"


@pytest.fixture(name="hermes_config")
def fixture_hermes_config(apps_dir: Path) -> Path:
    """Return the path to a shipping config for hermes"""
    return apps_dir / "shipping" / "hermes-deploy.yaml"


@pytest.fixture(name="scout_config")
def fixture_scout_config(apps_dir: Path) -> Path:
    """Return the path to a shipping config for scout"""
    return apps_dir / "shipping" / "scout-deploy.yaml"


@pytest.fixture(name="shipping_context")
def fixture_shipping_context(base_context: CGConfig, shipping_api: ShippingAPI) -> CGConfig:
    base_context.shipping_api_ = shipping_api
    return base_context
